#!/usr/bin/env python

from PyQt4.QtCore     import Qt, QByteArray, QIODevice, QModelIndex, QObject, \
                             QString, QStringList, QTextStream, QUrl, \
                             QVariant, SIGNAL
from PyQt4.QtGui      import QAbstractItemView, QAction, QComboBox, QDialog, \
                             QDialogButtonBox, QFileDialog, QGridLayout, \
                             QGroupBox, QHBoxLayout, QInputDialog, QLabel, \
                             QLineEdit, QListView, QMenuBar, QMessageBox, \
                             QProgressDialog, QPushButton, QStringListModel, \
                             QTableView, QToolButton, QVBoxLayout, QWidget
from PyQt4.QtSql      import QSqlQueryModel
from PyQt4.QtNetwork  import QNetworkAccessManager, QNetworkCookieJar, \
                             QNetworkReply, QNetworkRequest
from LTdatabase       import LTDatabase

version = '0.0.4'
exportUrl = QUrl('http://www.librarything.com/export-tab')
loginUrl  = QUrl('http://www.librarything.com/signup.php')

class LTException(Exception): #{{{1
    def __init__(self, source, error):
        self._source = source
        self._error  = error

    def __str__(self):
        return repr('%s: %s' % (self._source, self._error))

class LTFilterDefine(QDialog): #{{{1
    def __init__(self, parent, database): #{{{2
        QDialog.__init__(self, parent)

        self.setWindowTitle('Filters')
        self._database = database

        # Set up the list of filters {{{3
        self._filters  = QTableView()
        self._filters.setModel(QSqlQueryModel(self))
        self._filters.setCornerButtonEnabled(False)
        self._filters.verticalHeader().setVisible(False)
        self._filters.verticalHeader().setDefaultSectionSize(40)
        self._filters.horizontalHeader().setVisible(False)
        self._filters.setSelectionMode(QAbstractItemView.NoSelection)

        action = QAction('Remove Rule', self._filters)
        QObject.connect(action, SIGNAL('triggered(bool)'), self.removeRule)
        self._filters.addAction(action)
        self._filters.setContextMenuPolicy(Qt.ActionsContextMenu)

        # Set up the list of columns for filtering on {{{3
        self._colNames = database.getColumnNames()

        displayNamesTmp = self._colNames.keys()
        displayNames = QStringList()

        for col in displayNamesTmp:
            if self._database.hasContent(col):
                displayNames.append(col)

        displayNames.sort()

        self._filterFields = QComboBox()
        for val in displayNames:
            if self._colNames[val] == 'id':
                continue

            self._filterFields.addItem(val, QVariant(self._colNames[val]))

        # Set up the list of filtering methods {{{3
        self._filterMethods = QComboBox()
        self._filterMethods.addItem(QString('='))
        self._filterMethods.addItem(QString('contains'))
        self._filterMethods.addItem(QString('begins with'))

        # Set up the filter text field {{{3
        self._filterString = QLineEdit()

        # Set up the buttons {{{3
        btnLoad  = QPushButton('Load Filter')
        btnSave  = QPushButton('Save Filter')
        btnClear = QPushButton('Clear Rules')
        btnAdd   = QPushButton('Add Rule')

        btnAdd.setDefault(True)

        filterBtns = QDialogButtonBox(QDialogButtonBox.Ok|QDialogButtonBox.Cancel, Qt.Vertical)
        filterBtns.addButton(btnLoad, QDialogButtonBox.ActionRole)
        filterBtns.addButton(btnSave, QDialogButtonBox.ActionRole)
        filterBtns.addButton(btnClear, QDialogButtonBox.ResetRole)

        QObject.connect(filterBtns, SIGNAL('accepted()'), self.applyFilter)
        QObject.connect(filterBtns, SIGNAL('rejected()'), self.cancelFilter)
        QObject.connect(btnClear, SIGNAL('clicked(bool)'), self.clearFilter)
        QObject.connect(btnAdd, SIGNAL('clicked(bool)'), self.addRule)
        QObject.connect(btnLoad, SIGNAL('clicked(bool)'), self.loadFilter)
        QObject.connect(btnSave, SIGNAL('clicked(bool)'), self.saveFilter)

        # Layout all the elements {{{3
        hbox1 = QHBoxLayout()
        hbox1.addWidget(self._filterFields)
        hbox1.addWidget(self._filterMethods)
        hbox1.addWidget(self._filterString)
        hbox1.addWidget(btnAdd)

        hbox2 = QHBoxLayout()
        hbox2.addWidget(self._filters)
        hbox2.addWidget(filterBtns)

        vbox = QVBoxLayout()
        vbox.addLayout(hbox2)
        vbox.addLayout(hbox1)

        self.setLayout(vbox)

        # Load the data {{{3
        database.filterCopy('current', 'working')
        database.getFilter(self._filters.model(), 'working')

        self._filters.setColumnWidth(0, 200)
        self._filters.setColumnWidth(1, 100)
        self._filters.horizontalHeader().setStretchLastSection(True)

    def addRule(self, checked): #{{{2
        filterField = self._filterFields.itemData(self._filterFields.currentIndex())
        filterMethod = self._filterMethods.currentText()
        filterString = self._filterString.text()

        if filterString.isEmpty():
            return

        self._database.filterAddRule('working', filterField, filterMethod, filterString)
        self._database.getFilter(self._filters.model(), 'working')

    def removeRule(self, bool): #{{{2
        ruleRow = self._filters.currentIndex().row()

        if ruleRow == -1:
            return

        model = self._filters.model()

        filterField = self._colNames[model.index(ruleRow, 0).data().toString()]
        filterMethod = model.index(ruleRow, 1).data().toString()
        filterString = model.index(ruleRow, 2).data().toString()
        
        self._database.filterRemoveRule('working', filterField, filterMethod, filterString)
        self._database.getFilter(model, 'working')

    def applyFilter(self): #{{{2
        self._database.filterRename('working', 'current')
        self.accept()

    def cancelFilter(self): #{{{2
        self._database.filterDelete('working')
        self.reject()

    def clearFilter(self, checked): #{{{2
        self._database.filterDelete('working')
        self._database.getFilter(self._filters.model(), 'working')

    def loadFilter(self, checked): #{{{2
        name = QInputDialog.getItem(self, 'Load filter', 'Select filter to load', self._database.filterList())[0]

        if not name.isEmpty():
            self._database.filterCopy(name, 'working')
            self._database.getFilter(self._filters.model(), 'working')

    def saveFilter(self, checked): #{{{2
        if self._filters.model().rowCount() == 0:
            return

        name = QInputDialog.getText(self, 'Save filter', 'Enter name for filter')[0]

        if not name.isEmpty():
            self._database.filterCopy('working', name)

class LTColumnSelectGeneric(QDialog): #{{{1
    def __init__(self, parent, database): #{{{2
        QDialog.__init__(self, parent)
        self._database = database
        self.setMinimumSize(800, 480)

    def setupAvailableList(self, model): #{{{2
        self._availableColumns = QListView()
        self._availableColumns.setModel(model)
        self._availableColumns.setUniformItemSizes(True)
        self._availableColumns.setSelectionMode(QAbstractItemView.SingleSelection)
        QObject.connect(self._availableColumns, SIGNAL('doubleClicked(QModelIndex)'), self.moveItemClick)
        self._availableColumns.model().sort(0)

        lyt = QVBoxLayout()
        lyt.addWidget(QLabel('<b>Available fields</b>'))
        lyt.addWidget(self._availableColumns)

        return lyt

    def setupSelectedList(self, model, name): #{{{2
        self._selectedColumns = QListView()
        self._selectedColumns.setModel(model)
        self._selectedColumns.setUniformItemSizes(True)
        self._selectedColumns.setSelectionMode(QAbstractItemView.SingleSelection)
        QObject.connect(self._selectedColumns, SIGNAL('doubleClicked(QModelIndex)'), self.moveItemClick)
        self._selectedColumns.model() # This is needed to force the display

        lyt = QVBoxLayout()
        lyt.addWidget(QLabel('<b>' + name + '</b>'))
        lyt.addWidget(self._selectedColumns)

        return lyt

    def setupDialog(self, availLyt, selLyt): #{{{2
        # Set up the up/down arrows for ordering the display columns {{{3
        orderBtnUp = QToolButton()
        orderBtnUp.setArrowType(Qt.UpArrow)
        QObject.connect(orderBtnUp, SIGNAL('clicked(bool)'), self.orderItemUp)

        orderBtnDown = QToolButton()
        orderBtnDown.setArrowType(Qt.DownArrow)
        QObject.connect(orderBtnDown, SIGNAL('clicked(bool)'), self.orderItemDown)

        orderBtnBox = QVBoxLayout()
        orderBtnBox.addWidget(orderBtnUp)
        orderBtnBox.addWidget(orderBtnDown)

        # Set up the left/right arrows for moving to/from display {{{3
        selectBtnRight = QToolButton()
        selectBtnRight.setArrowType(Qt.RightArrow)
        QObject.connect(selectBtnRight, SIGNAL('clicked()'), self.moveItemRight)

        selectBtnLeft = QToolButton()
        selectBtnLeft.setArrowType(Qt.LeftArrow)
        QObject.connect(selectBtnLeft, SIGNAL('clicked()'), self.moveItemLeft)

        selectBtnBox = QVBoxLayout()
        selectBtnBox.addWidget(selectBtnRight)
        selectBtnBox.addWidget(selectBtnLeft)

        # Add the Save/Cancel buttons #{{{3
        exitButtons = QDialogButtonBox(QDialogButtonBox.Save|QDialogButtonBox.Cancel,Qt.Horizontal)
        QObject.connect(exitButtons, SIGNAL('accepted()'), self.saveSettings)
        QObject.connect(exitButtons, SIGNAL('rejected()'), self.cancelSelection)

        # Add all the widgets to the window {{{3
        selLyt.addWidget(exitButtons)
        lyt = QHBoxLayout()
        lyt.addLayout(availLyt)
        lyt.addLayout(selectBtnBox)
        lyt.addLayout(selLyt)
        lyt.addLayout(orderBtnBox)
        self.setLayout(lyt)

    def moveItemLeft(self): #{{{2
        item = self._selectedColumns.currentIndex()
        
        if item.row() == -1:
            return

        self.moveItem(item, self._selectedColumns, self._availableColumns)

    def moveItemRight(self): #{{{2
        item = self._availableColumns.currentIndex()

        if item.row() == -1:
            return

        self.moveItem(item, self._availableColumns, self._selectedColumns)

    def moveItemClick(self, item): #{{{2
        srcList = self.sender()
        dstList = None

        if srcList == self._availableColumns:
            dstList = self._selectedColumns
        elif srcList == self._selectedColumns:
            dstList = self._availableColumns
        else:
            print "Unknown caller"
            return

        self.moveItem(item, srcList, dstList)

    def orderItemUp(self, checked): #{{{2
        self.orderItem(-1)

    def orderItemDown(self, checked): #{{{2
        self.orderItem(1)

    def orderItem(self, offset): #{{{2
        selected = self._selectedColumns.currentIndex()
        selRow = selected.row()
        selColName = selected.data()

        if selRow == -1:
            return

        self._selectedColumns.model().removeRow(selRow)
        self._selectedColumns.model().insertRow(selRow + offset)
        newEntry = self._selectedColumns.model().index(selRow + offset)
        self._selectedColumns.model().setData(newEntry, selColName)
        self._selectedColumns.setCurrentIndex(newEntry)

    def moveItem(self, item, srcList, dstList): #{{{2
        colName = item.data()

        srcList.model().removeRow(item.row())

        newRow = dstList.model().rowCount()
        dstList.model().insertRow(newRow, QModelIndex())
        newItem = dstList.model().index(newRow)
        dstList.model().setData(newItem, colName)

        if dstList == self._availableColumns:
            dstList.model().sort(0)
            newItem = dstList.model().match(dstList.model().index(0), Qt.DisplayRole, QVariant(colName))[0]

        dstList.setCurrentIndex(newItem)

    def cancelSelection(self): #{{{2
        self.reject()

    def checkColumns(self, list): #{{{2
        new_list = QStringList()

        for col in list:
            if self._database.hasContent(col):
                new_list.append(col)

        return new_list

class LTColumnSelectDisplay(LTColumnSelectGeneric): #{{{1
    def __init__(self, parent, database): #{{{2
        LTColumnSelectGeneric.__init__(self, parent, database)

        self.setWindowTitle('Select and order columns to display')

        availableModel = QStringListModel(self.checkColumns(database.getNonDisplayColumns()))
        selectedModel  = QStringListModel(database.getDisplayColumns())

        availableList = self.setupAvailableList(availableModel)
        selectedList  = self.setupSelectedList(selectedModel, 'Displayed fields')

        availableModel.deleteLater()
        selectedModel.deleteLater()

        self.setupDialog(availableList, selectedList)

    def saveSettings(self): #{{{2
        self._database.saveDisplayOrder(self._selectedColumns.model().stringList())
        self.accept()

class LTColumnSelectSort(LTColumnSelectGeneric): #{{{1
    def __init__(self, parent, database): #{{{2
        LTColumnSelectGeneric.__init__(self, parent, database)

        self.setWindowTitle('Select and order columns to sort by')

        availableModel = QStringListModel(self.checkColumns(database.getNonSortColumns()))
        selectedModel  = QStringListModel(database.getSortColumns())

        availableList = self.setupAvailableList(availableModel)
        selectedList  = self.setupSelectedList(selectedModel, 'Sort fields')

        availableModel.deleteLater()
        selectedModel.deleteLater()

        self.setupDialog(availableList, selectedList)

    def saveSettings(self): #{{{2
        self._database.saveSortOrder(self._selectedColumns.model().stringList())
        self.accept()

class LTBookWindow(QDialog): #{{{1
    def __init__(self, parent, database, bookID): #{{{2
        QDialog.__init__(self, parent)

        self._bookInfo = database.getBookInfo(bookID)
        self._fieldNames = database.getFieldNames()

        self.setWindowTitle(self._bookInfo[QString('title')])

        lyt = QGridLayout(self)
        fldTitle = QLabel()
        fldTitle.setTextFormat(Qt.RichText)
        fldTitle.setText('<b>' + self.getBookInfo('title') + '</b>')

        fldAuthor = QLabel(self.getBookInfo('author_fl'))
        fldPublication = QLabel(self.getBookInfo('publication'))

        fldRating = self.createTaggedLabel('stars')
        fldTags = self.createTaggedLabel('tags')
        fldISBN = self.createTaggedLabel('isbn')

        lyt.addWidget(fldTitle, 1, 1, 1, 2)
        lyt.addWidget(fldAuthor, 2, 1, 1, 2)
        lyt.addWidget(fldPublication, 3, 1, 1, 2)
        lyt.addWidget(fldRating, 4, 1, 1, 1)
        lyt.addWidget(fldISBN, 4, 2, 1, 1)
        lyt.addWidget(fldTags, 5, 1, 1, 1)

        isbn = self.getBookInfo('isbn')
        if not isbn.isEmpty():
            fldLink = QLabel()
            fldLink.setOpenExternalLinks(True)
            fldLink.setTextFormat(Qt.RichText)
            fldLink.setText('<a href="http://www.librarything.com/isbn/' + self.getBookInfo('isbn') + '">Open in browser</a>')
            lyt.addWidget(fldLink, 6, 1, 1, 2, Qt.AlignHCenter)

    def getFieldName(self, colName): #{{{2
        return self._fieldNames[QString(colName)]

    def getBookInfo(self, colName): #{{{2
        return self._bookInfo[QString(colName)]

    def createTaggedLabel(self, field): #{{{2
        fldEntry = QLabel()
        fldEntry.setTextFormat(Qt.RichText)
        fldEntry.setText('<font color="brown">' + self.getFieldName(field) + ': </font>' + self.getBookInfo(field))

        return fldEntry

class LTAboutWindow(QDialog): #{{{1
    def __init__(self, parent): #{{{2
        QDialog.__init__(self, parent)

        self.setWindowTitle('About')

        aboutText  = '<h2>pybraryThing v' + version + '</h2>'
        aboutText += '<p>A <a href="http://www.librarything.com/">LibraryThing</a> export file viewer.</p>'
        aboutText += '<p>Written by Robin Hill <a href="mailto:maemo@robinhill.me.uk">&lt;maemo@robinhill.me.uk&gt;</a><br/>'
        aboutText += 'Please email or post to <a href="http://talk.maemo.org/">t.m.o</a> for support</p>'

        lyt = QHBoxLayout(self)
        txt = QLabel()
        txt.setOpenExternalLinks(True)
        txt.setTextFormat(Qt.RichText)
        txt.setText(aboutText)
        lyt.addWidget(txt)

class LTMainWindow(QWidget): #{{{1
    _maemo = False
    _downloadCompleteSet = False

    def __init__(self): #{{{2
        QWidget.__init__(self, None)

        try:
            from PyQt4.QtMaemo5 import QMaemo5InformationBox
            self._maemo = True
            self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, True)
        except:
            pass

        # Set to the N900 screen size to aid testing
        self.setMinimumSize(800, 480)
        self.setWindowTitle('LibraryThing Viewer')

        QHBoxLayout(self)

    def setupMenu(self): #{{{2
        menu = QMenuBar()

        # Set up the menu
        act = menu.addAction('Display Fields')
        QObject.connect(act, SIGNAL('triggered()'), self.displayFields)
        act = menu.addAction('Sort Order')
        QObject.connect(act, SIGNAL('triggered()'), self.sortOrder)
        act = menu.addAction('Filters')
        QObject.connect(act, SIGNAL('triggered()'), self.displayFilters)
        act = menu.addAction('Load Export File')
        QObject.connect(act, SIGNAL('triggered()'), self.loadFile)
        act = menu.addAction('Download Export File')
        QObject.connect(act, SIGNAL('triggered()'), self.downloadExportFile)
        act = menu.addAction('About')
        QObject.connect(act, SIGNAL('triggered()'), self.about)

        self.layout().setMenuBar(menu)
        self._menu = menu

    def setupScreen(self): #{{{2
        # Create the required objects
        self._tab      = QTableView()
        self._network  = QNetworkAccessManager(self)
        self._progress = QProgressDialog(self)
        self._tab.setModel(QSqlQueryModel(self))

        try:
            self._database = LTDatabase()
        except Exception, e:
            raise LTException('Error opening/creating database', str(e))

        # Set a cookie jar
        self._network.setCookieJar(QNetworkCookieJar())

        # Handle network responses
        QObject.connect(self._network, SIGNAL('finished(QNetworkReply *)'), self.replyFinished)

        # Set up the view pane
        self._tab.setCornerButtonEnabled(False)
        self._tab.setSelectionMode(QAbstractItemView.NoSelection)
        self._tab.verticalHeader().setVisible(False)
        self._tab.verticalHeader().setDefaultSectionSize(70)
        self._tab.horizontalHeader().setMovable(True)

        # Handle moving/resizing of columns
        QObject.connect(self._tab.horizontalHeader(), SIGNAL('sectionMoved(int, int, int)'), self.changeColumnOrder)
        QObject.connect(self._tab.horizontalHeader(), SIGNAL('sectionResized(int, int, int)'), self.changeColumnSize)
        QObject.connect(self._tab.horizontalHeader(), SIGNAL('sectionDoubleClicked(int)'), self._tab.resizeColumnToContents)

        # No way to get from the QVariant to a QAbstractKineticScroller
        # Fiddle with scrolling
#       scroller = self._tab.property('kineticScroller')

#       if scroller.isValid():
#           scroller.setLowFrictionEnabled(True)
#           scroller.setAxisLockThreshold('0.5')

        # Handle double-clicking an entry
        QObject.connect(self._tab, SIGNAL('doubleClicked(QModelIndex)'), self.viewBookInfo)
                    
        # Set up the main view
        self.layout().addWidget(self._tab)

        # Clear the progress indicator
        if self._maemo:
            self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, False)

    def loadData(self): #{{{2
        updateSpinner = False
        if self._maemo and not self.testAttribute(Qt.WA_Maemo5ShowProgressIndicator):
            updateSpinner = True
            self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, True)

        # Set the SQL query to use
        self._database.getQuery(self._tab.model())

        # Hide the ID column
        self._tab.setColumnHidden(0, True)

        # Set the column widths
        col_widths = self._database.getColumnWidths()
        view_size = self._tab.viewport().width()

        var_size_count = len(col_widths)
        for i in range(len(col_widths)):
            if col_widths[i] > 0:
                self._tab.setColumnWidth(i + 1, col_widths[i])
                view_size -= col_widths[i]
                var_size_count -= 1

        for i in range(len(col_widths)):
            if col_widths[i] == 0:
                width = 100
                if view_size > 0:
                    if var_size_count > 1:
                        width = int(view_size / var_size_count)
                        view_size -= width
                        var_size_count -= 1
                    else:
                        width = view_size

                self._tab.setColumnWidth(i + 1, width)
                self.changeColumnSize(i + 1, 0, width)

        if updateSpinner:
            self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, False)

    def displayFields(self): #{{{2
        winFields = LTColumnSelectDisplay(self, self._database)
        if winFields.exec_():
            self.loadData()

    def sortOrder(self): #{{{2
        winFields = LTColumnSelectSort(self, self._database)
        if winFields.exec_():
            self.loadData()

    def loadFile(self): #{{{2
        filename = QFileDialog.getOpenFileName(self, 'Open LT Export')

        if filename != None:
            if self._maemo:
                self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, True)
            self.setupProgressBar('Loading data')
            self._database.loadFile(filename, self._progress)
            self._tab.model().clear()
            self.loadData()
            if self._maemo:
                self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, False)

    def loginLT(self): #{{{2
        # TODO: Move the username/password to a config setting
        # Ask for user name & password
        userName = QInputDialog.getText(self, 'LibraryThing login', 'User name:', QLineEdit.Normal)[0]

        if userName.isEmpty():
            return

        password = QInputDialog.getText(self, 'LibraryThing login', 'Password:', QLineEdit.Password)[0]

        if password.isEmpty():
            return

        # Login
        post_data =  'formusername=%s&formpassword=%s&Submit=Sign+in' % (userName, password)
        reply = self._network.post(QNetworkRequest(loginUrl), QByteArray(post_data))

        self._loginBox = QDialog(self)
        self._loginBox.setWindowTitle('Logging in')
        lyt = QHBoxLayout(self._loginBox)
        lyt.addWidget(QLabel('Authenticating against the LibraryThing server'))
        self._loginBox.exec_()

    def downloadExportFile(self): #{{{2
        if self._downloadCompleteSet:
            QObject.disconnect(self, SIGNAL('downloadComplete()'), self.downloadExportFile)

        if self.haveLTCookie():
            if self._maemo:
                self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, True)

            reply = self._network.get(QNetworkRequest(exportUrl))

            self._loginBox = QDialog(self)
            self._loginBox.setWindowTitle('Downloading')
            lyt = QHBoxLayout(self._loginBox)
            lyt.addWidget(QLabel('Downloading LibraryThing export file'))
            self._loginBox.exec_()
        else:
            QObject.connect(self, SIGNAL('downloadComplete()'), self.downloadExportFile)
            self._downloadCompleteSet = True
            self.loginLT()

    def haveLTCookie(self): #{{{2
        for cookie in self._network.cookieJar().cookiesForUrl(exportUrl):
            if QString(cookie.name()) == QString('cookie_userid'):
                return True

        return False

    def setupProgressBar(self, title): #{{{2
        self._progress.setWindowTitle(title)
        self._progress.setMaximum(0)
        self._progress.setValue(0)

    def about(self): #{{{2
        aboutWin = LTAboutWindow(self)
        aboutWin.exec_()

    def changeColumnOrder(self, id, oldPos, newPos): #{{{2
        col = self._tab.model().record().fieldName(id)
        self._database.changeColumnOrder(col, oldPos, newPos)

    def changeColumnSize(self, id, oldSize, newSize): #{{{2
        col = self._tab.model().record().fieldName(id)
        self._database.changeColumnSize(col, newSize)

    def viewBookInfo(self, item): #{{{2
        bookID = self._tab.model().record(item.row()).value(0).toInt()[0]

        LTBookWindow(self, self._database, bookID).exec_()

    def replyFinished(self, reply = None): #{{{2
        if reply == None:
            reply = self.sender()

        if reply.error() == QNetworkReply.NoError:
            redir = reply.attribute(QNetworkRequest.RedirectionTargetAttribute).toUrl()
            if not redir.isEmpty():
                url = reply.request().url().resolved(redir)
                reply.deleteLater()

                req = QNetworkRequest(url)
                rep = self._network.get(req)
            elif reply.url() == exportUrl:
                self._loginBox.done(0)
                export = reply.readAll()
                reply.deleteLater()
                QObject.emit(self, SIGNAL('downloadComplete()'))

                stream = QTextStream(export, QIODevice.ReadOnly)

                self.setupProgressBar('Loading data')
                self._database.loadStream(stream, self._progress)
                self._tab.model().clear()
                self.loadData()

                if self._maemo:
                    self.setAttribute(Qt.WA_Maemo5ShowProgressIndicator, False)
            else:
                self._loginBox.done(0)
                res = reply.readAll()
                reply.deleteLater()

                if res.indexOf('Wrong user name or password') != -1:
                    QObject.disconnect(self, SIGNAL('downloadComplete()'), self.downloadExportFile)
                    self._downloadCompleteSet = False
                    QMessageBox.critical(self, 'Login failed', 'Username or password incorrect')

                QObject.emit(self, SIGNAL('downloadComplete()'))
        else:
            QMessageBox.critical(self, 'Network error', reply.errorString())
            reply.deleteLater()
            QObject.emit(self, SIGNAL('downloadComplete()'))

    def displayFilters(self): #{{{2
        filterWin = LTFilterDefine(self, self._database)
        if filterWin.exec_():
            self.loadData()
